import requests
import json
import socketio
from typing import Callable

class YTMD:
    namespace = "/api/v1/realtime"

    def __init__(self, app_id: str,
                app_name: str, 
                app_version: str,
                host: str = "127.0.0.1",
                port: int = 9863,
                token: str = None, logging=False):
        self.id = app_id
        self.name = app_name
        self.version = app_version
        self.host = host
        self.port = port
        self.token = token

        self.url = f"http://{self.host}:{self.port}/api/v1"

        self.session = requests.Session()
        self.session.headers.update({"Content-Type": "application/json"})

        self.socket = socketio.Client(logger=logging)
        self.registered_events = []

    def register_event(self, event: str, callback: Callable) -> None:
        """Register an event to be triggered when the event is emitted by YTMD

        Args:
            event (str): event from Events class
            callback (Callable): function to be called when event is triggered
        """
        if not callable(callback):
            raise Exception("Callback must be a function")
        
        if event in self.registered_events:
            raise Exception(f"Event {event} is already registered")

        self.socket.on(event, callback, namespace=self.namespace)
        self.registered_events.append(event)

    def connect(self) -> None:
        """Connect to YTMD Socket
        """
        self._check_token()

        if not self.registered_events:
            raise Exception("No events registered")
        
        try:
            self.socket.connect(
                f"ws://{self.host}:9863",
                auth={'token': self.token}, transports=["websocket"],
                namespaces=[self.namespace]
            )
        except Exception as e:
            raise Exception(f"Failed to connect to YTMD: {e}")

    def authenticate(self) -> str|None:
        """Authenticate with YTMD application and obtain token.
        If successful, the token is stored in the object and returned.
        This token should be saved and load them using update_token() method.
        """

        code = self.request_code()
        
        if code.status_code == 200:
            token = self.request_token(code.json()["code"])
            
            if token.status_code == 200:
                token = token.json()["token"]
                self.update_token(token)
                return token
            else:
                raise Exception(f"Failed to obtain token: {token.text}")
        
        raise Exception(f"Failed to obtain code: {code.text}")
    
    def update_token(self, token: str) -> None:
        """Whenever the object is created, this method should be
        used to register the token that was obtained from authenticate() method.
        """
        self.token = token
        self.session.headers.update({"Authorization": token})
    
    def request_code(self) -> requests.Response:
        """
        Internal method to request code from YTMD application
        """
        url = self.url + "/auth/requestcode"
        data = {
            "appId": self.id,
            "appName": self.name,
            "appVersion": self.version
        }
        return self.session.post(url, data=json.dumps(data))
    
    def request_token(self, code: str) -> requests.Response:
        """Internal method to request token from YTMD application

        Args:
            code (str): code that was obtained from request_code()
        """
        url = self.url + "/auth/request"
        data = {
            "appId": self.id,
            "code": code
        }
        return self.session.post(url, data=json.dumps(data))

    def _check_token(self) -> None:
        """Internal method to check if token is present in the object. If not, raise an exception.
        """
        if not self.token or not self.session.headers.get("Authorization"):
            raise Exception("Token is required to communicate with YTMD application. Please authenticate first.")
    
    def _command(self, command: str, data: int = None) -> requests.Response:
        """Internal method to send commands to the player"""
        self._check_token()

        url = self.url + "/command"
        request_data = { "command": command }
        
        if data:
            request_data["data"] = data

        return self.session.post(url, data=json.dumps(request_data))
    
    def get_version(self) -> list[str]:
        """Get the supported api versions of the player"""
        url = "http://" + f"{self.host}:{self.port}" + "/metadata"
        response = self.session.get(url)
        if response.status_code == 200:
            return response.json()['apiVersions']
        
        raise Exception(f"Failed to obtain metadata: {response.text}")

    def get_state(self) -> requests.Response:
        """
        Get the current state of the player
        """
        self._check_token()
        return self.session.get(self.url + "/state")
    
    def get_playlists(self) -> requests.Response:
        """
        Get the playlists of the player
        """
        self._check_token()
        return self.session.get(self.url + "/playlists")
    
    def toggle_playback(self) -> requests.Response:
        """
        Toggle playback
        """
        return self._command("playPause")
    
    def play(self) -> requests.Response:
        """
        Play the current track
        """
        return self._command("play")
    
    def pause(self) -> requests.Response:
        """
        pause the current track
        """
        return self._command("pause")
    
    def volume_up(self) -> requests.Response:
        """
        Increase the volume by 10%
        """
        return self._command("volumeUp")
    
    def volume_down(self) -> requests.Response:
        """
        Decrease the volume by 10%
        """
        return self._command("volumeDown")
    
    def set_volume(self, volume: int) -> requests.Response:
        """
        Set the volume to a specific value
        """
        return self._command("setVolume", volume)
    
    def mute(self) -> requests.Response:
        """
        Mute the player
        """
        return self._command("mute")
    
    def unmute(self) -> requests.Response:
        """
        Unmute the player
        """
        return self._command("unmute")
    
    def seek_to(self, time: int) -> requests.Response:
        """
        Seek to a specific time in the track
        """
        return self._command("seekTo", time)
    
    def next(self) -> requests.Response:
        """
        Play the next track in the queue
        """
        return self._command("next")
    
    def previous(self) -> requests.Response:
        """
        Play the previous track in the queue
        """
        return self._command("previous")
    
    def repeatMode(self, mode: int) -> requests.Response:
        """
        Set the repeat mode of the player

        Args:
            mode (int): 0 - None, 1 - All, 2 - One
        """
        return self._command("repeatMode", mode)
    
    def shuffle(self) -> requests.Response:
        """
        Shuffle the queue
        """
        return self._command("shuffle")
    
    def play_index(self, index: int) -> requests.Response:
        """
        Play a specific track in the queue
        """
        return self._command("playQueueIndex", index)
    
    def toggle_like(self) -> requests.Response:
        """
        Toggle the like status of the current track
        """
        return self._command("toggleLike")
    
    def toggle_dislike(self) -> requests.Response:
        """
        Toggle the dislike status of the current track
        """
        return self._command("toggleDislike")


from .player_state import PlayerState
from .video_state import VideoState

class Parser(object):
    player_state: PlayerState = None
    """player state"""

    video_state: VideoState = None
    """video state"""

    def __init__(self, data) -> None:
        self.player_state = PlayerState(data.get("player", {}))
        self.video_state = VideoState(data.get("video", {}))

    def __str__(self) -> str:
        return f"Player: {self.player_state}\nVideo: {self.video_state}"


import unittest
from ytmd_sdk import Parser
from ytmd_sdk.parser.queue import queueItem
from ytmd_sdk.parser.thumbnail import Thumbnail

class TestParser(unittest.TestCase):
    
    def test_parser(self):
        data = {}
        parser = Parser(data)
        self.assertEqual(parser.player_state.state, "Unknown")
        self.assertEqual(parser.player_state.video_progress, 0.0)
        self.assertEqual(parser.player_state.volume, 0)
        self.assertEqual(parser.player_state.muted, False)
        self.assertEqual(parser.player_state.adPlaying, False)
        self.assertEqual(parser.player_state.auto_play, False)
        self.assertEqual(parser.player_state.isGenerating, False)
        self.assertEqual(parser.player_state.isInfinite, False)
        self.assertEqual(parser.player_state.repeatMode, "Unknown")
        self.assertEqual(parser.player_state.selectedItemIndex, 0)
        self.assertEqual(parser.player_state.queue, [])

        self.assertEqual(parser.video_state.author, "")
        self.assertEqual(parser.video_state.channel_id, "")
        self.assertEqual(parser.video_state.title, "")
        self.assertEqual(parser.video_state.album, "")
        self.assertEqual(parser.video_state.album_id, "")
        self.assertEqual(parser.video_state.like_status, "Unknown")
        self.assertEqual(parser.video_state.duration_seconds, 0)
        self.assertEqual(parser.video_state.thumbnails, [])
        self.assertEqual(parser.video_state.id, "")

    def test_player_state(self):
        data = {
            "player": {
                "trackState": 1,
                "videoProgress": 13.2,
                "volume": 100,
                "muted": False,
                "adPlaying": True,
                "queue": {},
            }
        }
        parser = Parser(data)
        self.assertEqual(parser.player_state.state, "Playing")
        self.assertEqual(parser.player_state.video_progress, 13.2)
        self.assertEqual(parser.player_state.volume, 100)
        self.assertEqual(parser.player_state.muted, False)
        self.assertEqual(parser.player_state.adPlaying, True)
        self.assertEqual(parser.player_state.auto_play, False)
        self.assertEqual(parser.player_state.isGenerating, False)
        self.assertEqual(parser.player_state.isInfinite, False)
        self.assertEqual(parser.player_state.repeatMode, "Unknown")
        self.assertEqual(parser.player_state.selectedItemIndex, 0)
        self.assertEqual(parser.player_state.queue, [])

    def test_player_state_queue(self):
        data = {
            "player": {
                "queue": {
                    "autoplay": True,
                    "items": [
                        {
                            "thumbnails": [{"url": "url1", "width": 100, "height": 100},
                                           {"url": "url2", "width": 200, "height": 200},
                                           {"url": "url3", "width": 300, "height": 300}],
                            "title": "title",
                            "author": "author",
                            "duration": "3:30",
                            "selected": True,
                            "videoId": "videoId",
                            "counterparts": None
                        },
                        {
                            "thumbnails": [{"url": "url4", "width": 100, "height": 100},
                                           {"url": "url5", "width": 200, "height": 200},
                                           {"url": "url6", "width": 300, "height": 300}],
                            "title": "title",
                            "author": "author",
                            "duration": "1:30",
                            "selected": False,
                            "videoId": "videoId",
                            "counterparts": None
                        }
                    ]
                }
            }
        }
        parser = Parser(data)
        self.assertEqual(parser.player_state.auto_play, True)
        self.assertEqual(len(parser.player_state.queue), 2)
        self.assertIsInstance(parser.player_state.queue[0], queueItem)
        self.assertIsInstance(parser.player_state.queue[1], queueItem)

        self.assertIsInstance(parser.player_state.queue[0].thumbnails[0], Thumbnail)
        self.assertEqual(len(parser.player_state.queue[0].thumbnails), 3)

        self.assertEqual(parser.player_state.queue[0].thumbnails[0].url, "url1")
        self.assertEqual(parser.player_state.queue[0].thumbnails[0].width, 100)
        self.assertEqual(parser.player_state.queue[0].thumbnails[0].height, 100)
        self.assertEqual(parser.player_state.queue[0].title, "title")
        self.assertEqual(parser.player_state.queue[0].author, "author")
        self.assertEqual(parser.player_state.queue[0].duration, "3:30")
        self.assertEqual(parser.player_state.queue[0].selected, True)
        self.assertEqual(parser.player_state.queue[0].videoId, "videoId")
        self.assertEqual(parser.player_state.queue[0].counterparts, None)

    def test_video_state(self):
        data = {
            "video": {
                "author": "author",
                "channelId": "channelId",
                "title": "title",
                "album": "album",
                "albumId": "albumId",
                "likeStatus": 2,
                "thumbnails": [],
                "durationSeconds": 211,
                "id": "id"
            }
        }
        parser = Parser(data)
        self.assertEqual(parser.video_state.author, "author")
        self.assertEqual(parser.video_state.channel_id, "channelId")
        self.assertEqual(parser.video_state.title, "title")
        self.assertEqual(parser.video_state.album, "album")
        self.assertEqual(parser.video_state.album_id, "albumId")
        self.assertEqual(parser.video_state.like_status, "Like")
        self.assertEqual(parser.video_state.duration_seconds, 211)
        self.assertEqual(parser.video_state.thumbnails, [])
        self.assertEqual(parser.video_state.id, "id")

    def test_video_state_thumbnails(self):
        data = {
            "video": {
                "thumbnails": [{"url": "url1", "width": 100, "height": 100},
                               {"url": "url2", "width": 200, "height": 200},
                               {"url": "url3", "width": 300, "height": 300}],
            }
        }
        parser = Parser(data)
        self.assertEqual(len(parser.video_state.thumbnails), 3)
        self.assertIsInstance(parser.video_state.thumbnails[0], Thumbnail)
        self.assertIsInstance(parser.video_state.thumbnails[1], Thumbnail)
        self.assertIsInstance(parser.video_state.thumbnails[2], Thumbnail)

        self.assertEqual(parser.video_state.thumbnails[0].url, "url1")
        self.assertEqual(parser.video_state.thumbnails[0].width, 100)
        self.assertEqual(parser.video_state.thumbnails[0].height, 100)

        self.assertEqual(parser.video_state.thumbnails[1].url, "url2")
        self.assertEqual(parser.video_state.thumbnails[1].width, 200)
        self.assertEqual(parser.video_state.thumbnails[1].height, 200)

        self.assertEqual(parser.video_state.thumbnails[2].url, "url3")
        self.assertEqual(parser.video_state.thumbnails[2].width, 300)
        self.assertEqual(parser.video_state.thumbnails[2].height, 300)

if __name__ == '__main__':
    unittest.main()


from .converters import state_converter, repeatMode_converter
from .queue import queueItem
from typing import Annotated

class PlayerState(object):
    state: str = "Unknown"
    """player state"""
    video_progress: float = 0.0
    """video progress"""
    volume: int = 0
    """volume"""
    muted: bool = False
    """muted"""
    adPlaying: bool = False
    """ad playing"""
    queue: list[queueItem] = []
    """queue"""
    auto_play: bool = False
    """auto play"""
    isGenerating: bool = False
    """is generating"""
    isInfinite: bool = False
    """is infinite"""
    repeatMode: str = "Unknown"
    """repeat mode"""
    selectedItemIndex: int = 0
    """selected item index"""

    def __init__(self, data) -> None:
        self.state = state_converter.get(data.get("trackState", -1), "Unknown")
        self.video_progress = data.get("videoProgress", 0.0)
        self.volume = data.get("volume", 0)
        self.muted = data.get("muted", False)
        self.adPlaying = data.get("adPlaying", False)
        self.auto_play = data.get("queue", {}).get("autoplay", False)
        self.queue = [queueItem(item) for item in data.get("queue", {}).get("items", [])]
        self.isGenerating = data.get("isGenerating", False)
        self.isInfinite = data.get("isInfinite", False)
        self.repeatMode = repeatMode_converter.get(data.get("repeatMode", -1), "Unknown")
        self.selectedItemIndex = data.get("selectedItemIndex", 0)

    def __str__(self) -> str:
        return str({
                "State": self.state,
                "Video Progress": self.video_progress,
                "Volume": self.volume,
                "Muted": self.muted,
                "Ad Playing": self.adPlaying,
                "Auto Play": self.auto_play,
                "Queue": self.queue,
                "Is Generating": self.isGenerating,
                "Is Infinite": self.isInfinite,
                "Repeat Mode": self.repeatMode,
                "Selected Item Index": self.selectedItemIndex
              })


class Thumbnail(object):
    url: str = ""
    """url"""
    width: int = 0
    """width"""
    height: int = 0
    """height"""

    def __init__(self, data) -> None:
        self.url = data.get("url", "")
        self.width = data.get("width", 0)
        self.height = data.get("height", 0)

    def __str__(self) -> str:
        return str({"URL": self.url,"Width": self.width,"Height": self.height})


import unittest
from unittest.mock import patch
from ytmd_sdk import YTMD
from requests import Session

class TestYTMD(unittest.TestCase):
    def test_authenticate(self):
        with patch.object(Session, "post") as session_mock:
            ytmd = YTMD("touchportalytmd", "TouchPortalYTMD", "1.0.0")
            session_mock.return_value.status_code = 400
            session_mock.return_value.text = "Unittest"
            self.assertRaises(Exception, ytmd.authenticate)

    def test_get_version(self):
        with patch.object(Session, "get") as session_mock:
            ytmd = YTMD("touchportalytmd", "TouchPortalYTMD", "1.0.0")
            session_mock.return_value.status_code = 200
            session_mock.return_value.json.return_value = {"apiVersions": ["v1"]}
            response = ytmd.get_version()
            self.assertEqual(response, ["v1"])

    def test_get_state(self):
        with patch.object(Session, "get") as session_mock:
            ytmd = YTMD("touchportalytmd", "TouchPortalYTMD", "1.0.0")
            self.assertRaises(Exception, ytmd.get_state) # no token

            ytmd.update_token("token")
            session_mock.return_value.status_code = 200
            self.assertEqual(ytmd.get_state().status_code, 200)

    def test_ytmd_method(self):
        methods = ["get_playlists", "play", "pause", "volume_up", "volume_down", 
                   "set_volume", "mute", "unmute", "seek_to", "next", "previous", 
                   "repeatMode", "shuffle", "play_index", "toggle_like", "toggle_dislike"]
        
        for method in methods:
            with patch.object(Session, "post") as session_mock:
                print("Testing method: " + method)
                ytmd = YTMD("touchportalytmd", "TouchPortalYTMD", "1.0.0")

                if method in ["set_volume", "seek_to", "repeatMode", "play_index"]:
                    self.assertRaises(Exception, getattr(ytmd, method), 1)
                    ...
                else:
                    self.assertRaises(Exception, getattr(ytmd, method)) # no token

                ytmd.update_token("token")
                session_mock.return_value.status_code = 200

                if method in ["set_volume", "seek_to", "repeatMode", "play_index"]:
                    # response = getattr(ytmd, method)(1)
                    ...
                else:
                    response = getattr(ytmd, method)()
                    self.assertEqual(response.status_code, 200)



if __name__ == "__main__":
    unittest.main()


class Events:
    connect: str = "connect"
    disconnect: str = "disconnect"
    error: str = "error"
    state_update: str = "state-update"
    playlist_created: str = "playlist-created"
    playlist_deleted: str = "playlist-deleted"


from .thumbnail import Thumbnail
from typing import Annotated

class queueItem(object):
    title: str = ""
    """title"""
    author: str = ""
    """author"""
    duration: str = ""
    """duration"""
    thumbnails: list[Thumbnail] = []
    """thumbnails"""
    selected: bool = False
    """selected"""
    videoId: str = ""
    """video id"""
    counterparts: None
    """counterparts"""

    def __init__(self, data) -> None:
        self.title = data.get("title", "")
        self.author = data.get("author", "")
        self.duration = data.get("duration", "")
        self.thumbnails = [Thumbnail(item) for item in data['thumbnails']]
        self.selected = data.get("selected", False)
        self.videoId = data.get("videoId", "")
        self.counterparts = data.get("counterparts", None)

    def __str__(self) -> str:
        return str({"Title": self.title,"Author": self.author,"Duration": self.duration,"Thumbnails": self.thumbnails})


from .converters import like_status_converter
from .thumbnail import Thumbnail

class VideoState(object):
    author: str = ""
    """author"""
    channel_id: str = ""
    """channel id"""
    title: str = ""
    """title"""
    album: str = ""
    """album"""
    album_id: str = ""
    """album id"""
    like_status: int = 0
    """like status"""
    thumbnails: list[Thumbnail] = []
    """thumbnails"""
    duration_seconds: int = 0
    """duration seconds"""
    id: str = ""
    """id"""

    def __init__(self, data) -> None:
        self.author = data.get("author", "")
        self.channel_id = data.get("channelId", "")
        self.title = data.get("title", "")
        self.album = data.get("album", "")
        self.album_id = data.get("albumId", "")
        self.like_status = like_status_converter.get(data.get("likeStatus", -1), "Unknown")
        self.thumbnails = [Thumbnail(item) for item in data.get("thumbnails", [])]
        self.duration_seconds = data.get("durationSeconds", 0)
        self.id = data.get("id", "")


state_converter = {
    -1: "Unknown",
    0: "Paused",
    1: "Playing",
    2: "Buffering"
}

repeatMode_converter = {
    -1: "Unknown",
    0: "None",
    1: "All",
    2: "One"
}

like_status_converter = {
    -1: "Unknown",
    0: "Dislike",
    1: "Indifferent",
    2: "Like"
}